@@ -1,7 +1,7 @@
 /*
  *  ioping  -- simple I/0 latency measuring tool
  *
- *  Copyright (C) 2011-2013 Konstantin Khlebnikov <koct9i@gmail.com>
+ *  Copyright (C) 2011-2014 Konstantin Khlebnikov <koct9i@gmail.com>
  *
  *  This program is free software: you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -24,7 +24,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
-#include <inttypes.h>
 #include <getopt.h>
 #include <string.h>
 #include <unistd.h>
@@ -86,6 +85,14 @@
 # define HAVE_ERR_INCLUDE
 #endif
 
+#ifdef __sun__	/* Solaris */
+# include <sys/dkio.h>
+# include <sys/vtoc.h>
+# define HAVE_DIRECT_IO
+# define O_DIRECT	O_DSYNC
+# define HAVE_ERR_INCLUDE
+#endif
+
 #if defined(_POSIX_SYNCHRONIZED_IO) && _POSIX_SYNCHRONIZED_IO > 0
 # define HAVE_POSIX_FDATASYNC
 #endif
@@ -226,6 +233,7 @@ void usage(void)
 			"      -s <size>       request size (4k)\n"
 			"      -S <wsize>      working set size (1m)\n"
 			"      -o <offset>     working set offset (0)\n"
+			"      -k              keep and reuse temporary working file\n"
 			"      -L              use sequential operations (includes -s 256k)\n"
 			"      -A              use asynchronous I/O\n"
 			"      -C              use cached I/O\n"
@@ -342,14 +350,21 @@ long long parse_time(const char *str)
 	return parse_suffix(str, time_suffix);
 }
 
-void print_suffix(int64_t val, struct suffix *sfx)
+void print_suffix(long long val, struct suffix *sfx)
 {
+	int precision;
+
 	while (val < sfx->mul && sfx->mul > 1)
 		sfx++;
-	if (sfx->mul == 1)
-		printf("%" PRId64, val);
+
+	if (val % sfx->mul == 0)
+		precision = 0;
+	else if (val >= sfx->mul * 10)
+		precision = 1;
 	else
-		printf("%.1f", val * 1.0 / sfx->mul);
+		precision = 2;
+
+	printf("%.*f", precision, val * 1.0 / sfx->mul);
 	if (*sfx->txt)
 		printf(" %s", sfx->txt);
 }
@@ -408,11 +423,13 @@ ssize_t size = 1<<12;
 off_t wsize = 0;
 off_t temp_wsize = 1<<20;
 
+int keep_file = 0;
+
 off_t offset = 0;
 off_t woffset = 0;
 
 int request;
-int count = 0;
+int stop_at_request = 0;
 
 int exiting = 0;
 
@@ -425,7 +442,7 @@ void parse_options(int argc, char **argv)
 		exit(1);
 	}
 
-	while ((opt = getopt(argc, argv, "hvALRDCWBqi:w:s:S:c:o:p:P:")) != -1) {
+	while ((opt = getopt(argc, argv, "hvkALRDCWBqi:w:s:S:c:o:p:P:")) != -1) {
 		switch (opt) {
 			case 'h':
 				usage();
@@ -484,7 +501,10 @@ void parse_options(int argc, char **argv)
 				batch_mode = 1;
 				break;
 			case 'c':
-				count = parse_int(optarg);
+				stop_at_request = parse_int(optarg);
+				break;
+			case 'k':
+				keep_file = 1;
 				break;
 			case '?':
 				usage();
@@ -566,7 +586,7 @@ void parse_device(dev_t dev)
 
 #endif
 
-off_t get_device_size(int fd, struct stat *st)
+int get_device_size(int fd, struct stat *st)
 {
 	unsigned long long blksize = 0;
 	int ret = 0;
@@ -598,16 +618,21 @@ off_t get_device_size(int fd, struct stat *st)
 	part = label.d_partitions[DISKPART(st->st_rdev)];
 
 	blksize = DL_GETPSIZE(&part) * label.d_secsize;
+#elif defined(__sun__)
+	struct dk_minfo dkmp;
+
+	ret = ioctl(fd, DKIOCGMEDIAINFO, &dkmp);
+	blksize =  dkmp.dki_capacity * dkmp.dki_lbsize;
 #else
-# error no get disk size method
+# warning no get disk size method
+	ret = -1;
+	errno = ENOSYS;
+	blksize = 0;
 #endif
 	(void)fd;
-	(void)st;
 
-	if (ret)
-		err(2, "block get size ioctl failed");
-
-	return blksize;
+	st->st_size = blksize;
+	return ret;
 }
 
 ssize_t do_pwrite(int fd, void *buf, size_t nbytes, off_t offset)
@@ -730,26 +755,35 @@ static void aio_setup(void)
 
 static void aio_setup(void)
 {
-	errx(1, "asynchronous I/O not supportted by this platform");
+	errx(1, "asynchronous I/O not supported by this platform");
 }
 
 #endif
 
 #ifdef __MINGW32__
 
-int create_temp(char *path, char *template)
+int create_temp(char *path, char *name)
 {
-	int length = strlen(path) + strlen(template) + 2;
+	int length = strlen(path) + strlen(name) + 9;
 	char *temp = malloc(length);
+	DWORD action;
+	DWORD attr = 0;
 	HANDLE h;
-	DWORD attr;
 
 	if (!temp)
 		err(2, NULL);
-	snprintf(temp, length, "%s\\%s", path, template);
-	mktemp(temp);
 
-	attr = FILE_ATTRIBUTE_HIDDEN | FILE_FLAG_DELETE_ON_CLOSE;
+	snprintf(temp, length, "%s\\%s", path, name);
+
+	if (!keep_file) {
+		strcat(temp, ".XXXXXX");
+		mktemp(temp);
+		action = CREATE_NEW;
+		attr |= FILE_ATTRIBUTE_HIDDEN | FILE_FLAG_DELETE_ON_CLOSE;
+	} else {
+		action = OPEN_ALWAYS;
+	}
+
 	if (!cached)
 		attr |= FILE_FLAG_NO_BUFFERING | FILE_FLAG_WRITE_THROUGH;
 	if (randomize)
@@ -759,7 +793,7 @@ int create_temp(char *path, char *template)
 
 	h = CreateFile(temp, GENERIC_READ | GENERIC_WRITE,
 			FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
-			NULL, CREATE_ALWAYS, attr, NULL);
+			NULL, action, attr, NULL);
 
 	free(temp);
 	if (h == INVALID_HANDLE_VALUE)
@@ -787,26 +821,44 @@ void set_signal(void)
 
 #else /* __MINGW32__ */
 
-int create_temp(char *path, char *template)
+int create_temp(char *path, char *name)
 {
-	int length = strlen(path) + strlen(template) + 2;
+	int length = strlen(path) + strlen(name) + 9;
 	char *temp = malloc(length);
 	int fd;
 
 	if (!temp)
 		err(2, NULL);
-	snprintf(temp, length, "%s/%s", path, template);
 
+	snprintf(temp, length, "%s/%s", path, name);
+
+	if (keep_file) {
+		fd = open(temp, O_RDWR|O_CREAT, 0600);
+		if (fd < 0)
+			goto out;
+		goto done;
+	}
+
+#ifdef O_TMPFILE
+	fd = open(path, O_RDWR|O_TMPFILE, 0600);
+	if (fd >= 0)
+		goto done;
+#endif
+
+	strcat(temp, ".XXXXXX");
 	fd = mkstemp(temp);
 	if (fd < 0)
-		err(2, "failed to create temporary file at \"%s\"", path);
+		goto out;
 	if (unlink(temp))
 		err(2, "unlink \"%s\" failed", temp);
-	free(temp);
-# ifdef HAVE_DIRECT_IO
+
+done:
+#ifdef HAVE_DIRECT_IO
 	if (direct && fcntl(fd, F_SETFL, O_DIRECT))
 		errx(2, "fcntl failed, please retry without -D");
-# endif
+#endif
+out:
+	free(temp);
 	return fd;
 }
 
@@ -829,6 +881,26 @@ void set_signal(void)
 
 #endif /* __MINGW32__ */
 
+void random_memory(void *buf, size_t len)
+{
+	unsigned char *ptr = buf;
+	size_t i;
+
+	for (i = 0; i < len; i++)
+		ptr[i] = random();
+}
+
+/* not so random but much faster */
+void shake_memory(void *buf, size_t len)
+{
+	unsigned int *ptr = buf;
+	unsigned int word = random();
+	size_t i;
+
+	for (i = 0, len /= 4; i < len; i++)
+		ptr[i] ^= word;
+}
+
 int main (int argc, char **argv)
 {
 	long ret_size;
@@ -842,6 +914,8 @@ int main (int argc, char **argv)
 	double part_sum, part_sum2, part_mdev, part_avg;
 	long long time_now, time_next, period_deadline;
 
+	setvbuf(stdout, NULL, _IOLBF, 0);
+
 	parse_options(argc, argv);
 
 	interval_ts.tv_sec = interval / 1000000;
@@ -861,10 +935,11 @@ int main (int argc, char **argv)
 # if defined(HAVE_DIRECT_IO)
 	direct |= !cached;
 # else
-	if (!cached)
-		warnx("non-cached I/O not supportted by this platform,"
-				" results will be unreliable.");
-	cached = 1;
+	if (!cached && !write_test) {
+		warnx("non-cached read I/O not supported by this platform");
+		warnx("you can use write I/O to get reliable results");
+		cached = 1;
+	}
 # endif
 #endif
 
@@ -880,7 +955,7 @@ int main (int argc, char **argv)
 #ifdef HAVE_DIRECT_IO
 		flags |= O_DIRECT;
 #else
-		errx(1, "direct I/O not supportted by this platform");
+		errx(1, "direct I/O not supported by this platform");
 #endif
 
 #ifdef __MINGW32__
@@ -902,9 +977,23 @@ int main (int argc, char **argv)
 		if (fd < 0)
 			err(2, "failed to open \"%s\"", path);
 
-		device_size = get_device_size(fd, &st);
-		st.st_size = device_size;
-		fstype = "device";
+		if (get_device_size(fd, &st)) {
+			if (!S_ISCHR(st.st_mode))
+				err(2, "block get size ioctl failed");
+			st.st_size = offset + temp_wsize;
+			fstype = "character";
+			device = "device";
+		} else {
+			device_size = st.st_size;
+			fstype = "block";
+			device = "device ";
+		}
+
+		if (!cached && write_test && fdatasync(fd)) {
+			warnx("fdatasync not supported by \"%s\", "
+			      "enable cached requests", path);
+			cached = 1;
+		}
 	} else {
 		errx(2, "unsupported destination: \"%s\"", path);
 	}
@@ -921,14 +1010,28 @@ int main (int argc, char **argv)
 	ret = posix_memalign(&buf, 0x1000, size);
 	if (ret)
 		errx(2, "buffer allocation failed");
-	memset(buf, '*', size);
+
+	random_memory(buf, size);
 
 	if (S_ISDIR(st.st_mode)) {
-		fd = create_temp(path, "ioping.XXXXXX");
+		fd = create_temp(path, "ioping.tmp");
+		if (fd < 0)
+			err(2, "failed to create temporary file at \"%s\"", path);
+		if (keep_file) {
+			if (fstat(fd, &st))
+				err(2, "fstat at \"%s\" failed", path);
+			if (st.st_size >= offset + wsize)
+#ifndef __MINGW32__
+			    if (st.st_blocks >= (st.st_size + 511) / 512)
+#endif
+				goto skip_preparation;
+		}
 		for (woffset = 0 ; woffset + size <= wsize ; woffset += size) {
 			if (pwrite(fd, buf, size, offset + woffset) != size)
 				err(2, "write failed");
+			random_memory(buf, size);
 		}
+skip_preparation:
 		if (fsync(fd))
 			err(2, "fsync failed");
 	} else if (S_ISREG(st.st_mode)) {
@@ -986,6 +1089,9 @@ int main (int argc, char **argv)
 		}
 #endif
 
+		if (write_test)
+			shake_memory(buf, size);
+
 		this_time = now();
 
 		ret_size = make_request(fd, buf, size, offset + woffset);
@@ -1005,7 +1111,8 @@ int main (int argc, char **argv)
 
 		if (!quiet) {
 			print_size(ret_size);
-			printf(" from %s (%s %s", path, fstype, device);
+			printf(" %s %s (%s %s", write_test ? "to" : "from",
+					path, fstype, device);
 			if (device_size)
 				print_size(device_size);
 			printf("): request=%d time=", request);
@@ -1048,7 +1155,7 @@ int main (int argc, char **argv)
 		if (exiting)
 			break;
 
-		if (count && request >= count)
+		if (stop_at_request && request >= stop_at_request)
 			break;
 
 		if (deadline && time_next >= deadline)
@@ -1086,9 +1193,9 @@ int main (int argc, char **argv)
 		printf(" requests completed in ");
 		print_time(time_total);
 		printf(", ");
-		print_int(request * 1000000. / time_sum);
+		print_int(1000000. * request / time_sum);
 		printf(" iops, ");
-		print_size(request * size * 1000000. / time_sum);
+		print_size(1000000. * request * size / time_sum);
 		printf("/s\n");
 		printf("min/avg/max/mdev = ");
 		print_time(time_min);
